/*
 *      Copyright (C) 2000, The Xtal System, The University of
 *          Western Australia, xtal@crystal.uwa.edu.au.
 *
 * Distributed under the GNU General Public Licence as published by
 * the Free Software Foundation, in the hope that it will be useful
 * but WITHOUT ANY WARRANTY. See the GNU General Public License
 * for more details.
 *
 * Author:  D du Boulay ddb@R3401.msl.titech.ac.jp   
 *
 * This is incredibly incomplete experimental badly written 
 * Xtal-tcl/tk-OpenGL isosurface interface code derived loosely 
 * from the much more elegantly written and public domain:
 *
 *            Marching Cubes Example Program 
 *            by Cory Bloyd (dejaspaminacan@my-deja.com)
 *
 * A simple, portable and complete implementation of the Marching Cubes
 * and Marching Tetrahedrons algorithms in a single source file.
 * There are many ways that this code could be made faster, but the 
 * intent is for the code to be easy to understand.
 *
 * For a description of the algorithm go to
 * http://www.swin.edu.au/astronomy/pbourke/modelling/polygonise/
 *******************************************************************
 *
 *  For the record, OpenGL rendering works something like this:
 *
 *         MODELVIEW     PROJECTION    PERSPECTIVE           VIEWPORT
 *  Object -------> eye --------->clip --------->normailzing-------->window
 *  raw            world         +/-wc         +/-1 in x,y,z         coords
 *
 *  The MODELVIEW transformation matrix which is most important is 
 *  built up something like this (read backwards from right to left):
 *
 * [shift (0,0,-1)]x[user scale]x[user rotateY]x[user rotateX]x[shift]x
 * [scale to 1.0]x[orthog angstrom]*[scale to fractional coords]x
 *                             [assym unit grid translation]x(Map grid units)
 *
 * Basically this codebase creates on demand a data structure for each
 * required contour level of the xtal FOURR .map file, renders the
 * isosurface from the map using the marching cubes algorithm 
 * and draws the isosurface into a Togl tcl/tk extension OpenGL widget.
 * Specification of isosurface colours and lighting is made from 
 * within the asociated tcl/tk script.
 *
 * Think of it a proof of concept, and just forget that it has been 
 * proved much more elegantly by hundreds of other folk in the past.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include "togl.h"
//#include <GL/glu.h>
#include "tkxtal.h"

#include "stdio.h"
#include "math.h"

#ifdef __WIN32__
#else
#define __stdcall
#endif

/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */
#ifdef __SOLARIS__
extern int matherr();
int *tclDummyMathPtr = (int *) matherr;
#endif


typedef struct _GLvector
{
        GLfloat fX;
        GLfloat fY;
        GLfloat fZ;     
} GLvector ;

//These tables are used so that everything can be done in little loops that you can look at all at once
// rather than in pages and pages of unrolled code.

//a2fVertexOffset lists the positions, relative to vertex0, of each of the 8 vertices of a cube
static const GLfloat a2fVertexOffset[8][3] =
{
        {0.0, 0.0, 0.0},{1.0, 0.0, 0.0},{1.0, 1.0, 0.0},{0.0, 1.0, 0.0},
        {0.0, 0.0, 1.0},{1.0, 0.0, 1.0},{1.0, 1.0, 1.0},{0.0, 1.0, 1.0}
};

//a2iEdgeConnection lists the index of the endpoint vertices for each of the 12 edges of the cube
static const GLint a2iEdgeConnection[12][2] = 
{
        {0,1}, {1,2}, {2,3}, {3,0},
        {4,5}, {5,6}, {6,7}, {7,4},
        {0,4}, {1,5}, {2,6}, {3,7}
};

//a2fEdgeDirection lists the direction vector (vertex1-vertex0) for each edge in the cube
static const GLfloat a2fEdgeDirection[12][3] =
{
        {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
        {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
        {0.0, 0.0, 1.0},{0.0, 0.0, 1.0},{ 0.0, 0.0, 1.0},{0.0,  0.0, 1.0}
};

//a2iTetrahedronEdgeConnection lists the index of the endpoint vertices for each of the 6 edges of the tetrahedron
static const GLint a2iTetrahedronEdgeConnection[6][2] =
{
        {0,1},  {1,2},  {2,0},  {0,3},  {1,3},  {2,3}
};

//a2iTetrahedronEdgeConnection lists the index of verticies from a cube 
// that made up each of the six tetrahedrons within the cube
static const GLint a2iTetrahedronsInACube[6][4] =
{
        {0,5,1,6},
        {0,1,2,6},
        {0,2,3,6},
        {0,3,7,6},
        {0,7,4,6},
        {0,4,5,6},
};

static const GLfloat afAmbientWhite [] = {0.25, 0.25, 0.25, 0.40}; 
static const GLfloat afAmbientRed   [] = {0.25, 0.00, 0.00, 0.40}; 
static const GLfloat afAmbientGreen [] = {0.00, 0.25, 0.00, 0.40}; 
static const GLfloat afAmbientBlue  [] = {0.00, 0.00, 0.25, 0.40}; 
static const GLfloat afDiffuseWhite [] = {0.75, 0.75, 0.75, 0.40}; 
static const GLfloat afDiffuseRed   [] = {0.75, 0.00, 0.00, 0.40}; 
static const GLfloat afDiffuseGreen [] = {0.00, 0.75, 0.00, 0.40}; 
static const GLfloat afDiffuseBlue  [] = {0.00, 0.00, 0.75, 0.40}; 
static const GLfloat afSpecularWhite[] = {1.00, 1.00, 1.00, 0.40}; 
static const GLfloat afSpecularRed  [] = {1.00, 0.25, 0.25, 0.40}; 
static const GLfloat afSpecularGreen[] = {0.25, 1.00, 0.25, 0.40}; 
static const GLfloat afSpecularBlue [] = {0.25, 0.25, 1.00, 0.40}; 

#define true 1
#define false 0
typedef int bool; 



void vIdle();
void vDrawScene(); 
void vSpecial(int iKey, int iX, int iY);
void mouse(int button, int state, int x, int y) ;
void motion(int x, int y) ;

GLvoid vPrintHelp();
GLfloat fSample3(GLfloat fX, GLfloat fY, GLfloat fZ);
GLfloat (*fSample)(GLfloat fX, GLfloat fY, GLfloat fZ) = fSample3;

GLvoid vMarchingCubes();
GLvoid vMarchCube1(GLfloat fX, GLfloat fY, GLfloat fZ, GLfloat fScale);
GLvoid vMarchCube2(GLfloat fX, GLfloat fY, GLfloat fZ, GLfloat fScale);
GLvoid (*vMarchCube)(GLfloat fX, GLfloat fY, GLfloat fZ, GLfloat fScale) = vMarchCube1;

void check_error( char *where );
void create_cb( struct Togl *togl );
void display_cb(const struct Togl *togl );
void reshape_cb( struct Togl *togl );
void destroy_cb( struct Togl *togl );
int xrot_cb( struct Togl *togl,int argc, char *argv[]  );
int yrot_cb( struct Togl *togl,int argc, char *argv[] );
int scale_cb( struct Togl *togl,int argc, char *argv[] );

int cont_cb( struct Togl *togl, int argc, char *argv[] );
int aux_cb( struct Togl *togl, int argc, char *argv[] );
int light_cb( struct Togl *togl, int argc, char *argv[] );
int mindensity_cb( struct Togl *togl, int argc, char *argv[] );
int maxdensity_cb( struct Togl *togl, int argc, char *argv[] );
int epsfile_cb( struct Togl *togl, int argc, char *argv[] ) ;

static int   mapProperty=1;
static int   mapPoints[3]={0,0,0};
static int   mapOrigin[3]={0,0,0};
static int   gridPoints[3]={0,0,0};
static float   maxMapField[9]={0,0,0,0,0,0,0,0,0};
static float   minMapField[9]={0,0,0,0,0,0,0,0,0};
static GLdouble   orthog[16]={0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
static float   cdim[6]={0,0,0,0,0,0}; //a,b,c,cos a, cos b, cos g
static float  *xmap=NULL ;
static int    xdim=0 ;
static int    ydim=0 ;
static int    zdim=0 ;
static GLdouble  rotation[16]={1.,0.,0.,0., 0.,1.,0.,0., 0.,0.,1.,0., 0.,0.,0.,1.} ;

/* Viewing */
#define MAXGRID 63 
float sphi=0.0, stheta=0.0;
float sdepth = 5.0/4.0 * MAXGRID;
float zNear=15.0, zFar=100.0;
float aspect = 5.0/4.0;
long xsize, ysize;
int downX, downY;

static GLint     iDataSetSize = 16;
GLfloat   fStepSize = 1. ;
GLfloat   fTargetValue = 10.0;
GLvector  sSourcePoint[3];
GLboolean bLight = false;
GLenum    ePolygonMode = GL_FILL;
//GLenum    ePolygonMode = GL_LINE;

static GLfloat coord_scale = 1.0;
static GLfloat xrot = 0.0;
static GLfloat yrot = 0.0;
static GLfloat scale = 1.0;

static GLint width, height;
static GLboolean blend = GL_FALSE; 

bool leftButton = false, middleButton = false;

//float contours[]={20.,-20.,15.,-15.,10.,-10.,7.5,-7.5,5.0,-5.0} ;
float contours[]={20.,-20.,15.,-15.,10.,-10.} ;
//int maxLevel = 10;
int maxLevel = 6;

#define VALUERESET   1L<<0 
#define DISPCONTUR   1L<<1 
#define COLORENDER   1L<<2 
#define DIFFRENDER   1L<<3
#define AMBIRENDER   1L<<4
#define EMMIRENDER   1L<<5
#define SPECRENDER   1L<<6
#define SHINRENDER   1L<<7
#define FILLRENDER   1L<<8
#define LINERENDER   1L<<9

typedef struct _Contour {
	GLfloat   level;
	int       id;       /* need some consistent reference val for editing*/
	GLenum    polyMode;     /* = GL_FILL; */
	long      flags;
        GLfloat   ambient[4] ;
        GLfloat   diffuse[4] ;
        GLfloat   specular[4] ;
        GLfloat   emmision[4] ;
        GLfloat   colour[4] ;
        GLfloat   shininess ;
	GLuint    listID;        /* the GL list id */
}  Contour ;

typedef struct _setting {
	Contour  **levels;
	int      numlevs;
	int      asymUnit;
	int      cellUnit;
} setting;

setting settings;
setting *curSet=&settings ;

 
#if defined(WIN32)
EXTERN int              TkConsoleInit(Tcl_Interp *interp);
#endif /* WIN32 */

//EXPORT(int,Surfin_Init)( Tcl_Interp *interp );
int Surfin_Init(Tcl_Interp *interp);

/*
 *  Register module init function.
 *  We dont have an interpreter yet.
 *  That is created within tklib xcont_()
 *  but that subroutine takes a tcl script argument to initialize
 *  the graphics interface, and the tcl script needs to activate
 *  both the surfin module and the togl module within that script.
 *  Consequently we wind up here pre-preregistering some modules for static 
 *  loading in tklib xcont_()
 */

void sfpre_() {
     static const char *mod1 = "Togl" ;
     static const char *mod2 = "Surfin" ;

     xtalModPreRegister(mod1,Togl_Init,(Tcl_PackageInitProc *)NULL); 
     xtalModPreRegister(mod2,Surfin_Init,(Tcl_PackageInitProc *)NULL);

}

 
//int Surfin_Init( Tcl_Interp *interp )   /*   init from tcl script */
int Surfin_Init(Tcl_Interp *interp)   /*   init from tcl script */
{
   /*
    * Specify the C callback functions for widget creation, display,
    * and reshape.
    */
   Togl_CreateFunc( create_cb );
   Togl_DisplayFunc( display_cb );
   Togl_ReshapeFunc( reshape_cb );
   Togl_DestroyFunc( destroy_cb );
 
   /*
    * Make a new Togl widget command so the Tcl code can set a C variable.
    */
   Togl_CreateCommand( "xrot", xrot_cb );
   Togl_CreateCommand( "yrot", yrot_cb );
   Togl_CreateCommand( "scale", scale_cb );
   Togl_CreateCommand( "epsfile", epsfile_cb );

   /*
    * Maybe these should be independent of the Togle command
    */
   Togl_CreateCommand( "mindensity", mindensity_cb );
   Togl_CreateCommand( "maxdensity", maxdensity_cb );

   /*
    * functions to set isosurface contour levels
    */
   Togl_CreateCommand( "cont",      cont_cb );
   Togl_CreateCommand( "aux",     aux_cb );
   Togl_CreateCommand( "light",     light_cb );
 
   /*
    * Call Tcl_CreateCommand for application-specific commands, if
    * they weren't already created by the init procedures called above.
    */
 
   return TCL_OK;
}



void c3dini_(int *prop, int mappts[], int maporg[], int ngs[], float max[],float min[],float fparams[],float *data)  /* init from fortran */
{ 
        int argc=1 ;  
	int i;
	GLdouble sing,h1 ;
        GLfloat afPropertiesAmbient [] = {0.50, 0.50, 0.50, 1.00}; 
        GLfloat afPropertiesDiffuse [] = {0.75, 0.75, 0.75, 1.00}; 
        GLfloat afPropertiesSpecular[] = {1.00, 1.00, 1.00, 1.00}; 
        GLfloat afPropertiesPosition[] = {0.00, 0.00, 5.00, 1.00}; 

	xmap = data;              // save a pointer to fortran data 
	mapProperty = *prop ;
	for(i=0;i<3;i++) {
	    mapPoints[i] = mappts[i] ;
	    mapOrigin[i] = maporg[i] ;
	    gridPoints[i] = ngs[i] ;
	}
	for(i=0;i<mapProperty;i++) {
		maxMapField[i]=max[i];
		minMapField[i]=min[i];
	}
	for(i=0;i<6;i++) {
		cdim[i]= fparams[i];
	}

	sing = sqrt(1.-cdim[5]*cdim[5]); //          Sin(gamma)      
        h1=(cdim[3]-cdim[4]*cdim[5])/sing;//(cos(a)-cos(b)*cos(g))/sin(g)
	orthog[0]=cdim[0];//                     A              
	orthog[4]=cdim[1]*cdim[5];//             B*cos(g)        
	orthog[8]=cdim[2]*cdim[4];//             C*cos(b)         
	orthog[5]=cdim[1]*sing ; //                B*sin(g)           
	orthog[9]=cdim[2]*h1;   //              C*[h1]              
	orthog[10]=cdim[2]*sqrt(1.-cdim[4]*cdim[4]-h1*h1);
	orthog[15]=1.0;
	

        glClearColor( 0.0, 0.0, 0.0, 1.0 ); 
        glClearDepth( 1.0 ); 

        glEnable(GL_DEPTH_TEST); 
	//glEnable(GL_LIGHTING);
        glDisable(GL_LIGHTING);
        glPolygonMode(GL_FRONT_AND_BACK, ePolygonMode);

        glEnable( GL_LIGHT0 ); 
        glLightfv( GL_LIGHT0, GL_AMBIENT,  afPropertiesAmbient); 
        glLightfv( GL_LIGHT0, GL_DIFFUSE,  afPropertiesDiffuse); 
        glLightfv( GL_LIGHT0, GL_SPECULAR, afPropertiesSpecular); 
//        glLightfv( GL_LIGHT0, GL_POSITION, afPropertiesPosition); 
        glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0); 
        //glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 0.0); 
       // glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmbient) ;

        // Nice Perspective Correction
//        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); 
        glEnable(GL_COLOR_MATERIAL ); 

        //glMaterialfv(GL_BACK,  GL_AMBIENT,   afAmbientGreen); 
        //glMaterialfv(GL_BACK,  GL_DIFFUSE,   afDiffuseGreen); 
        //glMaterialfv(GL_FRONT, GL_AMBIENT,   afAmbientBlue); 
        //glMaterialfv(GL_FRONT, GL_DIFFUSE,   afDiffuseBlue); 
        //glMaterialfv(GL_FRONT, GL_SPECULAR,  afSpecularWhite); 
        //glMaterialf( GL_FRONT, GL_SHININESS, 25.0); 

	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	curSet->asymUnit=1;
	curSet->cellUnit=1;
	curSet->numlevs=0;

	// ideally would like to alert Tcl side that data is registered
	// by activating a tcl script here. But we have no interpreter!
}



int cont_cb( struct Togl *togl, int argc, char *argv[] ) {
   int aptr=2;
   int newcon=0;  /* assume a previous contour */
   int id = 0;
   int level=-1 ;
   char string[100];
   Contour  *contour ;
   Tcl_Interp *interp = Togl_Interp(togl);
   int levels = curSet->numlevs+1;

   if (argc < 3) {
      Tcl_SetResult( interp,
           "wrong # args: should be \"pathName cont ?ID/new?  \"",
           TCL_STATIC );
      return TCL_ERROR;
   }

   if (strcmp(argv[aptr],"new")==0) {
      // alloc mem for a new contour
      curSet->levels=(Contour**)realloc(curSet->levels,
	    				levels*sizeof(Contour **));
      if (curSet->levels==NULL ||
          (contour=(Contour *)malloc((unsigned)(sizeof(Contour)))) == NULL) {
          Tcl_SetResult( interp,
           "\"pathName newContour ?value?\": contour memory allocation error",
           TCL_STATIC );
          return TCL_ERROR;
      }
      (curSet->numlevs)++;
      contour->level = 0.0 ;
      contour->id  = curSet->numlevs;   // nothing associated with it yet
      contour->flags = VALUERESET |COLORENDER | EMMIRENDER | AMBIRENDER | 
	                DIFFRENDER | SPECRENDER | FILLRENDER | SHINRENDER ;
      contour->ambient[0] = 0.5 ;       // grey colour
      contour->ambient[1] = 0.5 ;
      contour->ambient[2] = 0.5 ;
      contour->ambient[3] = 0.5 ;
      contour->diffuse[0] = 0.5 ;       // grey colour
      contour->diffuse[1] = 0.5 ;
      contour->diffuse[2] = 0.5 ;
      contour->diffuse[3] = 0.5 ;
      contour->emmision[0] = 0.5 ;       // grey colour
      contour->emmision[1] = 0.5 ;
      contour->emmision[2] = 0.5 ;
      contour->emmision[3] = 0.5 ;
      contour->specular[0] = 0.5 ;       // grey colour
      contour->specular[1] = 0.5 ;
      contour->specular[2] = 0.5 ;
      contour->specular[3] = 0.5 ;
      contour->colour[0] = 0.5 ;       // grey colour
      contour->colour[1] = 0.5 ;
      contour->colour[2] = 0.5 ;
      contour->colour[3] = 0.5 ;
      contour->listID  =  curSet->numlevs ; // >0
      level = curSet->numlevs - 1 ; 
      curSet->levels[level]= contour ; // append at end of list
      newcon=1;
   } 
   else if (strcmp(argv[aptr],"delete")==0) {
/*	printf("destroying old contours\n"); */
	/* destroy individual contour levels */
	for (level=0 ;level <curSet->numlevs;level++) {
		contour= curSet->levels[level] ;
		if(contour == NULL) continue ; // already deleted?
                glDeleteLists((GLuint)contour->listID,(GLsizei)1);
/*		printf("delete %d \n",level); fflush(stdout); */
		free(contour) ;
	}
	/* destroy  the list of contour levels */
	if (curSet->levels) free(curSet->levels) ;
        curSet->numlevs=0 ; 
	/* destroy the accumulated list of settings (for backtracking)???*/

        strcpy( interp->result,"0");
        return TCL_OK;
   }
   else {    // look for an old contour match
      id  = atoi(argv[aptr] ) ;
      for (level=0;level<curSet->numlevs;level++) {
          if (curSet->levels[level]->id ==id) break;
      }
      if (level>curSet->numlevs || curSet->levels[level]->id !=id) {
         Tcl_SetResult( interp,
                  "unregistered id in \"pathName cont ?id?\"",
                   TCL_STATIC );
         return TCL_ERROR;
      }
      contour=curSet->levels[level] ;  // save reference to the old contour
   }
   aptr++; 

   // now its a free for all to decide what is being set
   while(aptr<argc-1){
      if (strcmp(argv[aptr],"value")==0) {
	 if (argc-aptr-2>0) {
              Tcl_SetResult( interp,
                  "no \"val\" in \"pathName cont ?id? value ?val?\"",
                   TCL_STATIC );
              return TCL_ERROR;
         }
         contour->level  = atof(argv[aptr+1] ) ;
	 if (!newcon) {      //delete exist disp list
             glDeleteLists(contour->listID,(GLsizei)1);
	 }
         contour->flags |= VALUERESET ;  // re-evaluate the polyhedra
	 aptr+=2 ;
      }
      else if (strcmp(argv[aptr],"delete")==0) {
         glDeleteLists(contour->listID,(GLsizei)1);
         free(contour) ;
         curSet->levels[level] = NULL ;
	 break;
      }
      else if (strcmp(argv[aptr],"color")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName cont ?id? color ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->colour[0]= atof(argv[aptr+1] ) ;
         contour->colour[1]= atof(argv[aptr+2] ) ;
         contour->colour[2]= atof(argv[aptr+3] ) ;
         contour->colour[3]= atof(argv[aptr+4] ) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"ambience")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName cont ?id? ambience ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->ambient[0]= atof(argv[aptr+1] ) ;
         contour->ambient[1]= atof(argv[aptr+2] ) ;
         contour->ambient[2]= atof(argv[aptr+3] ) ;
         contour->ambient[3]= atof(argv[aptr+4] ) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"diffuse")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName cont ?id? diffuse ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->diffuse[0]= atof(argv[aptr+1] ) ;
         contour->diffuse[1]= atof(argv[aptr+2] ) ;
         contour->diffuse[2]= atof(argv[aptr+3] ) ;
         contour->diffuse[3]= atof(argv[aptr+4] ) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"specular")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName cont ?id? specular ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->specular[0]= atof(argv[aptr+1] ) ;
         contour->specular[1]= atof(argv[aptr+2] ) ;
         contour->specular[2]= atof(argv[aptr+3] ) ;
         contour->specular[3]= atof(argv[aptr+4] ) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"emmisive")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName cont ?id? emmisive ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->emmision[0]= atof(argv[aptr+1] ) ;
         contour->emmision[1]= atof(argv[aptr+2] ) ;
         contour->emmision[2]= atof(argv[aptr+3] ) ;
         contour->emmision[3]= atof(argv[aptr+4] ) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"shininess")==0) {
	 if (argc-aptr-2>0) {
              Tcl_SetResult( interp,
                  " is \"pathName cont ?id? shininess ?0.<=val<=1.?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->shininess= atof(argv[aptr+1] ) ;
	 aptr+=2 ;
      }
      else if (strcmp(argv[aptr],"flags")==0) {
	 if (argc-aptr-2>0) {
              Tcl_SetResult( interp,
                  " is \"pathName cont ?id? flags ?flag?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         contour->flags = atoi(argv[aptr+1] ) ;
	 aptr+=2 ;
      }
      else if (strcmp(argv[aptr],"poly")==0) {
         if        (strcmp(argv[aptr+1],"fill")==0) {
	    contour->polyMode= GL_FILL ;
	 } else if (strcmp(argv[aptr+1],"line")==0) {
	    contour->polyMode= GL_LINE ;
	 } else {
            Tcl_SetResult( interp,
            "unexpected argument in \"pathName cont ?id? poly ?fill/line? \"",
              TCL_STATIC );
            return TCL_ERROR;
	 }
	 aptr+=2 ;
      }
      else {
         Tcl_SetResult( interp,
           "unexpected argument in \"pathName cont ?id? ...\"",
              TCL_STATIC );
         return TCL_ERROR;
      }

   }



   /* Let result string equal ID */
   sprintf(string,"%d",contour->id);
   strcpy( interp->result, string );
   Togl_PostRedisplay(togl);
   return TCL_OK;
}



/*
 * Switch off unit cell outline ? 
 * Switch off aux cell outline
 * Set contour resolution
 * Set  background colour?
 *
 */
int aux_cb( struct Togl *togl, int argc, char *argv[] ) {
   int aptr;
   Tcl_Interp *interp = Togl_Interp(togl);
   return TCL_OK;
}



int light_cb( struct Togl *togl, int argc, char *argv[] ) {
   int aptr;
   int id,j;
   Tcl_Interp *interp = Togl_Interp(togl);
   char num[] ={'0','1','2','3','4','5','6','7','8','9'};
   GLfloat globalAmbient[] = {0.5f, 0.5f, 0.5f, 1.0f} ;
   GLfloat  val;
   GLfloat  vec[4];
   GLfloat  col[4];
   GLenum   light;
   char string[10];
   /* 
    * set up the lighting
    * Can have 8 light sources
    * should control all of them
    */
   if (argc < 3) {
      Tcl_SetResult( interp,
           "wrong # args: should be \"pathName light ?N/enable/disable?  \"",
           TCL_STATIC );
      return TCL_ERROR;
   }
   aptr=2;
   if         (strcmp(argv[aptr],"enable")==0) {
      glEnable(GL_LIGHTING);
      return TCL_OK;
   } else if (strcmp(argv[aptr],"disable")==0) {
      glDisable(GL_LIGHTING);
      return TCL_OK;
   } else if (strcmp(argv[aptr],"model")==0) {
      if (strcmp(argv[aptr+1],"ambience")==0) {
          glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmbient) ; 
      }
      else if (strcmp(argv[aptr+1],"localview")==0) {
         glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0) ;
      }
      else if (strcmp(argv[aptr+1],"twoside")==0) {
          glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0); 
      }
      return TCL_OK;
   } else  {
      for(id=0;id<9;id++) {
         if (*(argv[aptr])==num[id]) break;
      }
      if(id<8) id = atoi(argv[aptr] ) ;
      switch (id) {
       case  0:  light=GL_LIGHT0;break;
       case  1:  light=GL_LIGHT1;break;
       case  2:  light=GL_LIGHT2;break;
       case  3:  light=GL_LIGHT3;break;
       case  4:  light=GL_LIGHT4;break;
       case  5:  light=GL_LIGHT5;break;
       case  6:  light=GL_LIGHT6;break;
       case  7:  light=GL_LIGHT7;break;
       default:
          Tcl_SetResult( interp,
           "unknown args: should be \"pathName light ?0<=N<=8?  \"",
           TCL_STATIC );
          return TCL_ERROR;
      }
   }
   aptr++;

   // now its a free for all to decide what is being set
   while(aptr<argc-1){

      if (strcmp(argv[aptr],"enable")==0) {
         glEnable(light);
	 aptr++ ;
      }
      else if (strcmp(argv[aptr],"disable")==0) {
         glDisable(light);
	 aptr++ ;
      }
      else if (strcmp(argv[aptr],"position")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName light ?N? position ?x? ?y? ?z?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         vec[0]= atof(argv[aptr+1] ) ;
         vec[1]= atof(argv[aptr+2] ) ;
         vec[2]= atof(argv[aptr+3] ) ;
         vec[3]= atof(argv[aptr+4] ) ;
         glLightfv(light, GL_POSITION,vec) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"direction")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName light ?N? direction ?x? ?y? ?z?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         vec[0]= atof(argv[aptr+1] ) ;
         vec[1]= atof(argv[aptr+2] ) ;
         vec[2]= atof(argv[aptr+3] ) ;
         vec[3]= atof(argv[aptr+4] ) ;
         glLightfv(light, GL_SPOT_DIRECTION,vec) ;
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"cutoff")==0) {
	 if (argc-aptr-2>0) {
              Tcl_SetResult( interp,
                  " is \"pathName light ?N? cutoff ?val?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         val=  atof(argv[aptr+1] ) ;
         glLightf(light, GL_SPOT_CUTOFF, val);
	 aptr+=2 ;
      }
      else if (strcmp(argv[aptr],"exponent")==0) {
	 if (argc-aptr-2>0) {
              Tcl_SetResult( interp,
                  " is \"pathName light ?N? exponent ?val?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         val=  atof(argv[aptr+1] ) ;
         glLightf(light, GL_SPOT_EXPONENT, val);
	 aptr+=2 ;
      }
      else if (strcmp(argv[aptr],"ambience")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName light ?N? ambience ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         col[0]= atof(argv[aptr+1] ) ;
         col[1]= atof(argv[aptr+2] ) ;
         col[2]= atof(argv[aptr+3] ) ;
         col[3]= atof(argv[aptr+4] ) ;
         glLightfv(light, GL_AMBIENT, col);
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"diffuse")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName light ?N? diffuse ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         col[0]= atof(argv[aptr+1] ) ;
         col[1]= atof(argv[aptr+2] ) ;
         col[2]= atof(argv[aptr+3] ) ;
         col[3]= atof(argv[aptr+4] ) ;
         glLightfv(light, GL_DIFFUSE, col);
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"specular")==0) {
	 if (argc-aptr-5>0) {
              Tcl_SetResult( interp,
                  "command is \"pathName light ?N? specular ?R? ?G? ?B? ?O?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         col[0]= atof(argv[aptr+1] ) ;
         col[1]= atof(argv[aptr+2] ) ;
         col[2]= atof(argv[aptr+3] ) ;
         col[3]= atof(argv[aptr+4] ) ;
         glLightfv(light, GL_SPECULAR, col);
	 aptr+=5 ;
      }
      else if (strcmp(argv[aptr],"flags")==0) {
	 if (argc-aptr-2>0) {
              Tcl_SetResult( interp,
                  " is \"pathName light ?N? flags ?flag?\"",
                   TCL_STATIC );
              return TCL_ERROR;
	 }
         j = atoi(argv[aptr+1] ) ;  //ignore any flags
	 aptr+=2 ;
      }
      else {
         Tcl_SetResult( interp,
           "unexpected argument in \"pathName light ?N? ...\"",
              TCL_STATIC );
/*         printf("%s\n",argv[aptr]); */
         return TCL_ERROR;
      }
   }


   /* Let result string equal ID */
   sprintf(string,"%d",id);
   strcpy( interp->result, string );
   Togl_PostRedisplay(togl);
   return TCL_OK;
}


/*
 *  render graphics to a filename
 */
int epsfile_cb( struct Togl *togl, int argc, char *argv[] )
{
   int rgbFlag=0 ;
   Tcl_Interp *interp = Togl_Interp(togl);
 
   /* error checking */
   if (argc <3 || argc >4) {
      Tcl_SetResult( interp,
               "wrong # args: should be \"pathName epsfile ?filename? ?rgb? \"",
                TCL_STATIC );
      return TCL_ERROR;
   }
   if (argc ==4 && strncmp(argv[3],"rgb",3)==0) {rgbFlag=1;}

   return Togl_DumpToEpsFile(togl,argv[2], rgbFlag, display_cb ) ;
}

/*
 *  return minimum density array  (1 ,3 or 9 elements )
 */

int mindensity_cb( struct Togl *togl, int argc, char *argv[] )
{
   Tcl_Interp *interp = Togl_Interp(togl);
   char *format= "list %e %e %e %e %e %e %e %e %e\n";
   char result[200] = " ";
 
   /* error checking */
   if (argc !=2 ) {
      Tcl_SetResult( interp,
               "wrong # args: should be \"pathName mindensity \"",
                TCL_STATIC );
      return TCL_ERROR;
   }
   sprintf(result,format,minMapField[0],minMapField[1],minMapField[2] \
		   ,minMapField[3] ,minMapField[4] ,minMapField[5] \
		   ,minMapField[6] ,minMapField[7] ,minMapField[8] ) ;
   return Tcl_Eval(interp,result) ;
}

/*
 *  return maximum density array  (1 ,3 or 9 elements
 */
int maxdensity_cb( struct Togl *togl, int argc, char *argv[] )
{
   Tcl_Interp *interp = Togl_Interp(togl);
   char *format= "list %e %e %e %e %e %e %e %e %e\n";
   char result[200] = " ";
 
   /* error checking */
   if (argc !=2 ) {
      Tcl_SetResult( interp,
               "wrong # args: should be \"pathName maxdensity \"",
                TCL_STATIC );
      return TCL_ERROR;
   }
   sprintf(result,format,maxMapField[0],maxMapField[1],maxMapField[2]
		   ,maxMapField[3] ,maxMapField[4] ,maxMapField[5]
		   ,maxMapField[6] ,maxMapField[7] ,maxMapField[8] ) ;
   return Tcl_Eval(interp,result) ;
}

sortContours() {
	int i,j;
	Contour *lev;
	for(i=0;i<curSet->numlevs-1;i++) {
		for(j=i+1;j<curSet->numlevs;j++) {
			if( curSet->levels[i] == NULL&&curSet->levels[j]);//swap
			else if( curSet->levels[j] == NULL) continue ;
			else if( fabs(curSet->levels[i]->level)  >=
			         fabs(curSet->levels[j]->level)) continue ;
			lev =curSet->levels[i];
			curSet->levels[i]= curSet->levels[j];
			curSet->levels[j] = lev ;
		}
        }
}

void create_cb( struct Togl *togl ) {
}


/* 
 * destroy settings on close - free mem
 */
void destroy_cb( struct Togl *togl ) {
	char *argv[3]={"dummy","cont","delete"};
	cont_cb(togl,3,argv) ; // destroy contours for curSett
}


/*
 * Togl widget reshape callback.  This is called by Tcl/Tk when the widget
 * has been resized.  Typically, we call glViewport and perhaps setup the
 * projection matrix.
 */

void reshape_cb( struct Togl *togl ) {
   	GLint iWidth = Togl_Width( togl );
   	GLint iHeight = Togl_Height( togl );

        GLfloat fAspect, fHalfWorldSize = (1.4142135623730950488016887242097/2); 
        glViewport( 0, 0, iWidth, iHeight ); 
        glMatrixMode (GL_PROJECTION);
        glLoadIdentity ();

        if(iWidth <= iHeight)
        {
                fAspect = (GLfloat)iHeight / (GLfloat)iWidth;
                glOrtho(-fHalfWorldSize, fHalfWorldSize, -fHalfWorldSize*fAspect,
                        fHalfWorldSize*fAspect, -10*fHalfWorldSize, 10*fHalfWorldSize);
        }
        else
        {
                fAspect = (GLfloat)iWidth / (GLfloat)iHeight; 
                glOrtho(-fHalfWorldSize*fAspect, fHalfWorldSize*fAspect, -fHalfWorldSize,
                        fHalfWorldSize, -10*fHalfWorldSize, 10*fHalfWorldSize);
        }
 
        glMatrixMode( GL_MODELVIEW );
}



void check_error( char *where )
{
   GLenum error;
 
   while (1) {
      error = glGetError();
      if (error==GL_NO_ERROR) {
         break;
      }
//      printf("OpenGL error near %s: %s\n", where, gluErrorString( error ) );
   }
}


void display_cb(const struct Togl *togl ) 
{ 
	int level ;
        int newSettings;
	int i;
	float d[3],o[3],c[3];
        float scaleFac;
	GLfloat *v;
	Contour *contour;

	check_error("begin display\n");
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); 


	// could draw some permanent background image here

        glPushMatrix(); 
	glLoadIdentity();      //begin with identity matrix
	// load small rotations identity
        glRotatef(    yrot, 0.0, 1.0, 0.0);
        glRotatef(xrot, 1.0, 0.0, 0.0);
	yrot=0.0; xrot=0.0; //reset
	// apply small rotations to current setting
	glMultMatrixd(rotation);
	glGetDoublev(GL_MODELVIEW_MATRIX,rotation);  //save for next time

	//start again
	glLoadIdentity();      //begin with identity matrix
        glTranslatef(0.0, 0.0, -5.0); // back out  
	glMultMatrixd(rotation);          // apply users rotation

        glScalef( scale, scale, scale );  // apply users scale


        glPushMatrix(); 
	
        scaleFac=0.0;
	for(i=0;i<3;i++) {
	  o[i]=(float)mapOrigin[i]/(float)gridPoints[i];
	  d[i]=((float)mapPoints[i]-1.)/(float)gridPoints[i] + o[i];
          c[i]=-(d[i]+o[i])*0.5;
          if (scaleFac<cdim[i]) scaleFac=cdim[i];
	}
        if (scale<=0.0)scaleFac=1. ;
        else           scaleFac=1./scaleFac;
        
        //scale orthog angstroms to unity
	glScalef(scaleFac,scaleFac,scaleFac); 
	glMultMatrixd(orthog); // convert fract to angstrom

        //working in unit coords - origin is midpoint os asym unit 
        glTranslatef(c[0], c[1], c[2]); //translate to map centre
        //working in unit coords - origin is 0 0 0

        glPushAttrib(GL_LIGHTING_BIT); 
                glDisable(GL_LIGHTING);
                glColor3f(1.0, 1.0, 1.0);
		if (curSet->asymUnit) {// draw asym unit outline
        		glBegin(GL_LINES);
			glVertex3f(o[0],o[1],o[2]); glVertex3f(d[0],o[1],o[2]);
			glVertex3f(o[0],o[1],o[2]); glVertex3f(o[0],d[1],o[2]);
			glVertex3f(o[0],o[1],o[2]); glVertex3f(o[0],o[1],d[2]);

			glVertex3f(d[0],o[1],o[2]); glVertex3f(d[0],d[1],o[2]);
			glVertex3f(d[0],o[1],o[2]); glVertex3f(d[0],o[1],d[2]);
			glVertex3f(o[0],d[1],o[2]); glVertex3f(d[0],d[1],o[2]);
			glVertex3f(o[0],d[1],o[2]); glVertex3f(o[0],d[1],d[2]);
			glVertex3f(o[0],o[1],d[2]); glVertex3f(d[0],o[1],d[2]);
			glVertex3f(o[0],o[1],d[2]); glVertex3f(o[0],d[1],d[2]);

			glVertex3f(d[0],d[1],d[2]); glVertex3f(o[0],d[1],d[2]);
			glVertex3f(d[0],d[1],d[2]); glVertex3f(d[0],o[1],d[2]);
			glVertex3f(d[0],d[1],d[2]); glVertex3f(d[0],d[1],o[2]);

        		glEnd();

		}
		if (curSet->cellUnit) {// draw unit cell outline
        		glBegin(GL_LINES);
			glVertex3f(0.0,0.0,0.0); glVertex3f(1.0,0.0,0.0);
			glVertex3f(0.0,0.0,0.0); glVertex3f(0.0,1.0,0.0);
			glVertex3f(0.0,0.0,0.0); glVertex3f(0.0,0.0,1.0);

			glVertex3f(1.0,0.0,0.0); glVertex3f(1.0,1.0,0.0);
			glVertex3f(1.0,0.0,0.0); glVertex3f(1.0,0.0,1.0);
			glVertex3f(0.0,1.0,0.0); glVertex3f(1.0,1.0,0.0);
			glVertex3f(0.0,1.0,0.0); glVertex3f(0.0,1.0,1.0);
			glVertex3f(0.0,0.0,1.0); glVertex3f(1.0,0.0,1.0);
			glVertex3f(0.0,0.0,1.0); glVertex3f(0.0,1.0,1.0);

			glVertex3f(1.0,1.0,1.0); glVertex3f(1.0,1.0,0.0);
			glVertex3f(1.0,1.0,1.0); glVertex3f(0.0,1.0,1.0);
			glVertex3f(1.0,1.0,1.0); glVertex3f(1.0,0.0,1.0);
        		glEnd();
			
		}
        glPopAttrib(); 



        //normalize grid units to 1.0
	glScalef(1./gridPoints[0],1./gridPoints[1],1./gridPoints[2]); 

        //translate to map origin in grid units
        glTranslatef((float)mapOrigin[0],(float)mapOrigin[1],
					(float)mapOrigin[2]);

	newSettings=0 ;
	for (level=0;level<curSet->numlevs;level++) {
	   contour = curSet->levels[level];
	   if ( contour->flags & VALUERESET) {
	      printf("%d  level %f \n",level, contour->level);

	      glNewList((GLuint)curSet->levels[level]->listID, GL_COMPILE) ;
       	      glBegin(GL_TRIANGLES);
	      fTargetValue = contour->level ;
              vMarchingCubes();
              glEnd();
	      glEndList() ;
	      newSettings=1 ;
	      contour->flags ^= VALUERESET ;
	   }
	}
	if (newSettings) sortContours();  // highest lev first
	for (level=0;level<curSet->numlevs;level++) {
	   contour = curSet->levels[level];
	      if (!(contour->flags &  DISPCONTUR) ) continue ; 
       	         glMaterialfv(GL_BACK,  GL_AMBIENT,   afAmbientGreen); 
	         glMaterialfv(GL_BACK,  GL_DIFFUSE,   afDiffuseGreen); 
 		 if (contour->flags & EMMIRENDER) glMaterialfv(GL_FRONT, 
				      GL_EMISSION, contour->emmision);
 		 if (contour->flags & AMBIRENDER) glMaterialfv(GL_FRONT, 
			              GL_AMBIENT, contour->ambient);
 		 if (contour->flags & DIFFRENDER) glMaterialfv(GL_FRONT, 
				      GL_DIFFUSE, contour->diffuse);
 		 if (contour->flags & SPECRENDER) glMaterialfv(GL_FRONT, 
				      GL_SPECULAR, contour->specular);
 		 if (contour->flags & COLORENDER) 
				      glColor4fv(contour->colour);  
 		 if (contour->flags & SHINRENDER) glMaterialf( GL_FRONT, 
				      GL_SHININESS, contour->shininess);
		if (contour->flags & FILLRENDER) {
		      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
		      glCallList(contour->listID);
		}
		if (!(contour->flags & FILLRENDER) && 
	             (contour->flags & LINERENDER)    ) {
		      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		      glCallList(contour->listID);
		      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
		}
		if ( (contour->flags & FILLRENDER) && 
	             (contour->flags & LINERENDER)    ) {
                      glPushAttrib(GL_LIGHTING_BIT); 
		      glDisable(GL_LIGHTING);
		      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		      glEnable(GL_POLYGON_OFFSET_LINE);
		      //glPolygonOffset(m_PolygonOffset,-1.0f);
		      glCallList(contour->listID);
		      glDisable(GL_POLYGON_OFFSET_LINE);
		      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
                      glPopAttrib(); 
		}

	}

        glPopMatrix(); // back to simple rotated coords 
        glPopMatrix(); 

	Togl_SwapBuffers( togl );
	check_error("end display\n");
}

int xrot_cb( struct Togl *togl, int argc, char *argv[] )
{
   Tcl_Interp *interp = Togl_Interp(togl);
    
      /* error checking */
      if (argc != 3) {
            Tcl_SetResult( interp,
			                         "wrong # args: should be \"pathName setXrot ?angle?\"",
						                      TCL_STATIC );
	          return TCL_ERROR;
		     }
      
        xrot = atof( argv[2] );
  
    Togl_PostRedisplay(togl);
     
       /* Let result string equal value */
       strcpy( interp->result, argv[2] );
     return TCL_OK;
}

int yrot_cb( struct Togl *togl, int argc, char *argv[] )
{
	   Tcl_Interp *interp = Togl_Interp(togl);
	    
	      /* error checking */
	      if (argc != 3) {
		            Tcl_SetResult( interp,
					                         "wrong # args: should be \"pathName setYrot ?angle?\"",
								                      TCL_STATIC );
			          return TCL_ERROR;
				     }
	       
	         yrot = atof( argv[2] );
		  
		    Togl_PostRedisplay(togl);
		     
		       /* Let result string equal value */
		       strcpy( interp->result, argv[2] );
		          return TCL_OK;
}

int scale_cb( struct Togl *togl, int argc, char *argv[] )
{
	   Tcl_Interp *interp = Togl_Interp(togl);
	    
	      /* error checking */
	      if (argc != 3) {
		            Tcl_SetResult( interp,
					                         "wrong # args: should be \"pathName scale ?value?\"",
								                      TCL_STATIC );
			          return TCL_ERROR;
				     }
	       
	         scale = atof( argv[2] );
		  
		    Togl_PostRedisplay(togl);
		     
		       /* Let result string equal value */
		       strcpy( interp->result, argv[2] );
		          return TCL_OK;
}

//fGetOffset finds the approximate point of intersection of the surface
// between two points with the values fValue1 and fValue2
GLfloat fGetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired)
{
        GLdouble fDelta = fValue2 - fValue1;

        if(fDelta == 0.0)
        {
                return 0.5;
        }
        return (fValueDesired - fValue1)/fDelta;
}


//vGetColor generates a color from a given position and normal of a point
GLvoid vGetColor(GLvector *rfColor, GLvector *rfPosition, GLvector *rfNormal)
{
	if (fSample==fSample3) {
		if (fTargetValue<0.) {
        		rfColor->fX = 1.00 ;
		        rfColor->fY = 0.00 ;
      			rfColor->fZ = 0.00 ;
		} else {
        		rfColor->fX = 0.00 ;
		        rfColor->fY = 0.00 ;
      			rfColor->fZ = 1.00 ;
		}
	}
	else {
        GLfloat fX = rfNormal->fX;
        GLfloat fY = rfNormal->fY;
        GLfloat fZ = rfNormal->fZ;
        rfColor->fX = (fX > 0.0 ? fX : 0.0) + (fY < 0.0 ? -0.5*fY : 0.0) + (fZ < 0.0 ? -0.5*fZ : 0.0);
        rfColor->fY = (fY > 0.0 ? fY : 0.0) + (fZ < 0.0 ? -0.5*fZ : 0.0) + (fX < 0.0 ? -0.5*fX : 0.0);
        rfColor->fZ = (fZ > 0.0 ? fZ : 0.0) + (fX < 0.0 ? -0.5*fX : 0.0) + (fY < 0.0 ? -0.5*fY : 0.0);
	}
}

GLvoid vNormalizeVector(GLvector *rfVectorResult, GLvector *rfVectorSource)
{
        GLfloat fOldLength;
        GLfloat fScale;

        fOldLength = sqrtf( (rfVectorSource->fX * rfVectorSource->fX) +
                            (rfVectorSource->fY * rfVectorSource->fY) +
                            (rfVectorSource->fZ * rfVectorSource->fZ) );

        if(fOldLength == 0.0)
        {
                rfVectorResult->fX = rfVectorSource->fX;
                rfVectorResult->fY = rfVectorSource->fY;
                rfVectorResult->fZ = rfVectorSource->fZ;
        }
        else
        {
                fScale = 1.0/fOldLength;
                rfVectorResult->fX = rfVectorSource->fX*fScale;
                rfVectorResult->fY = rfVectorSource->fY*fScale;
                rfVectorResult->fZ = rfVectorSource->fZ*fScale;
        }
}



GLfloat fSample3(GLfloat fX, GLfloat fY, GLfloat fZ)
{
	// crude
        //int ptr = ((int)fZ) * (xdim*ydim) + ((int)fY)*xdim + (int) fX ;
        long ptr = ((long)(fZ > 0.0 ? fZ : 0.0)) * (mapPoints[0]*mapPoints[1]) + 
		  ((long)(fY> 0.0 ? fY : 0.0)) *  mapPoints[0] + 
		   (long)(fX> 0.0 ? fX : 0.0) ;

        GLfloat fResult = (GLdouble) xmap[ptr] ;
	// printf("%f %f %f    %f \n",fX,fY,fZ,fResult);
        return fResult;
}



//vGetNormal() finds the gradient of the scalar field at a point
//This gradient can be used as a very accurate vertx normal for lighting calculations
GLvoid vGetNormal(GLvector *rfNormal, GLfloat fX, GLfloat fY, GLfloat fZ)
{
        //rfNormal->fX = fSample(fX-0.01, fY, fZ) - fSample(fX+0.01, fY, fZ);
        //rfNormal->fY = fSample(fX, fY-0.01, fZ) - fSample(fX, fY+0.01, fZ);
        //rfNormal->fZ = fSample(fX, fY, fZ-0.01) - fSample(fX, fY, fZ+0.01);
        rfNormal->fX = fSample(fX-1., fY, fZ) - fSample(fX+1., fY, fZ);
        rfNormal->fY = fSample(fX, fY-1., fZ) - fSample(fX, fY+1., fZ);
        rfNormal->fZ = fSample(fX, fY, fZ-1.) - fSample(fX, fY, fZ+1.);
	if (fTargetValue >0.0) { 
        rfNormal->fX = -rfNormal->fX ;
        rfNormal->fY = -rfNormal->fY ;
        rfNormal->fZ = -rfNormal->fZ ;
	}
        vNormalizeVector(rfNormal, rfNormal);
}


//vMarchCube1 performs the Marching Cubes algorithm on a single cube
GLvoid vMarchCube1(GLfloat fX, GLfloat fY, GLfloat fZ, GLfloat fScale)
{
        extern GLint aiCubeEdgeFlags[256];
        extern GLint a2iTriangleConnectionTable[256][16];

        GLint iCorner, iVertex, iVertexTest, iEdge, iTriangle, iFlagIndex, iEdgeFlags;
        GLfloat fOffset;
        GLvector sColor;
        GLfloat afCubeValue[8];
        GLvector asEdgeVertex[12];
        GLvector asEdgeNorm[12];

        //Make a local copy of the values at the cube's corners
        for(iVertex = 0; iVertex < 8; iVertex++)
        {
          afCubeValue[iVertex]=fSample(fX + a2fVertexOffset[iVertex][0]*fScale,
                                       fY + a2fVertexOffset[iVertex][1]*fScale,
                                       fZ + a2fVertexOffset[iVertex][2]*fScale);
        }

        //Find which vertices are inside of the surface and which are outside
        iFlagIndex = 0;
        for(iVertexTest = 0; iVertexTest < 8; iVertexTest++)
        {
                if(afCubeValue[iVertexTest] <= fTargetValue) 
                        iFlagIndex |= 1<<iVertexTest;
        }

        //Find which edges are intersected by the surface
        iEdgeFlags = aiCubeEdgeFlags[iFlagIndex];

        //If the cube is entirely inside or outside of the surface, then there will be no intersections
        if(iEdgeFlags == 0) 
        {
                return;
        }

        //Find the point of intersection of the surface with each edge
        //Then find the normal to the surface at those points
        for(iEdge = 0; iEdge < 12; iEdge++)
        {
                //if there is an intersection on this edge
                if(iEdgeFlags & (1<<iEdge))
                {
                        fOffset = fGetOffset(
				afCubeValue[ a2iEdgeConnection[iEdge][0] ], 
                                afCubeValue[ a2iEdgeConnection[iEdge][1] ], 
				   fTargetValue);

                        asEdgeVertex[iEdge].fX = fX + (
				a2fVertexOffset[a2iEdgeConnection[iEdge][0]][0]+
			      	fOffset * a2fEdgeDirection[iEdge][0]) * fScale;
                        asEdgeVertex[iEdge].fY = fY + (
				a2fVertexOffset[a2iEdgeConnection[iEdge][0]][1]+
			      	fOffset * a2fEdgeDirection[iEdge][1]) * fScale;
                        asEdgeVertex[iEdge].fZ = fZ + (
				a2fVertexOffset[a2iEdgeConnection[iEdge][0]][2]+
			      	fOffset * a2fEdgeDirection[iEdge][2]) * fScale;

                        vGetNormal(&asEdgeNorm[iEdge], asEdgeVertex[iEdge].fX, 
				asEdgeVertex[iEdge].fY, asEdgeVertex[iEdge].fZ);
                }
        }


        //Draw the triangles that were found.  There can be up to five per cube
        for(iTriangle = 0; iTriangle < 5; iTriangle++)
        {
                if(a2iTriangleConnectionTable[iFlagIndex][3*iTriangle] < 0)
                        break;

                for(iCorner = 0; iCorner < 3; iCorner++)
                {
                        iVertex = a2iTriangleConnectionTable[iFlagIndex][3*iTriangle+iCorner];

                        vGetColor(&sColor, &asEdgeVertex[iVertex], &asEdgeNorm[iVertex]);
                        //glColor4f(sColor.fX, sColor.fY, sColor.fZ,fabs(fTargetValue/contours[0]));
                        //glColor3f(sColor.fX, sColor.fY, sColor.fZ,0.5 );
                        glNormal3f(asEdgeNorm[iVertex].fX,   asEdgeNorm[iVertex].fY,   asEdgeNorm[iVertex].fZ);
                        glVertex3f(asEdgeVertex[iVertex].fX, asEdgeVertex[iVertex].fY, asEdgeVertex[iVertex].fZ);
                }
        }
}

//vMarchTetrahedron performs the Marching Tetrahedrons algorithm on a single tetrahedron
GLvoid vMarchTetrahedron(GLvector *pasTetrahedronPosition, GLfloat *pafTetrahedronValue)
{
        extern GLint aiTetrahedronEdgeFlags[16];
        extern GLint a2iTetrahedronTriangles[16][7];

        GLint iEdge, iVert0, iVert1, iEdgeFlags, iTriangle, iCorner, iVertex, iFlagIndex = 0;
        GLfloat fOffset, fInvOffset, fValue = 0.0;
        GLvector asEdgeVertex[6];
        GLvector asEdgeNorm[6];
        GLvector sColor;

        //Find which vertices are inside of the surface and which are outside
        for(iVertex = 0; iVertex < 4; iVertex++)
        {
                if(pafTetrahedronValue[iVertex] <= fTargetValue) 
                        iFlagIndex |= 1<<iVertex;
        }

        //Find which edges are intersected by the surface
        iEdgeFlags = aiTetrahedronEdgeFlags[iFlagIndex];

        //If the tetrahedron is entirely inside or outside of the surface, then there will be no intersections
        if(iEdgeFlags == 0)
        {
                return;
        }
        //Find the point of intersection of the surface with each edge
        // Then find the normal to the surface at those points
        for(iEdge = 0; iEdge < 6; iEdge++)
        {
                //if there is an intersection on this edge
                if(iEdgeFlags & (1<<iEdge))
                {
                        iVert0 = a2iTetrahedronEdgeConnection[iEdge][0];
                        iVert1 = a2iTetrahedronEdgeConnection[iEdge][1];
                        fOffset = fGetOffset(pafTetrahedronValue[iVert0], pafTetrahedronValue[iVert1], fTargetValue);
                        fInvOffset = 1.0 - fOffset;

                        asEdgeVertex[iEdge].fX = fInvOffset*pasTetrahedronPosition[iVert0].fX  +  fOffset*pasTetrahedronPosition[iVert1].fX;
                        asEdgeVertex[iEdge].fY = fInvOffset*pasTetrahedronPosition[iVert0].fY  +  fOffset*pasTetrahedronPosition[iVert1].fY;
                        asEdgeVertex[iEdge].fZ = fInvOffset*pasTetrahedronPosition[iVert0].fZ  +  fOffset*pasTetrahedronPosition[iVert1].fZ;
                        
                        vGetNormal(&asEdgeNorm[iEdge], asEdgeVertex[iEdge].fX, asEdgeVertex[iEdge].fY, asEdgeVertex[iEdge].fZ);
                }
        }
        //Draw the triangles that were found.  There can be up to 2 per tetrahedron
        for(iTriangle = 0; iTriangle < 2; iTriangle++)
        {
                if(a2iTetrahedronTriangles[iFlagIndex][3*iTriangle] < 0)
                        break;

                for(iCorner = 0; iCorner < 3; iCorner++)
                {
                        iVertex = a2iTetrahedronTriangles[iFlagIndex][3*iTriangle+iCorner];

                        vGetColor(&sColor, &asEdgeVertex[iVertex], &asEdgeNorm[iVertex]);
                        glColor4f(sColor.fX, sColor.fY, sColor.fZ,fabs(fTargetValue/contours[0]));
                        //glColor3f(sColor.fX, sColor.fY, sColor.fZ);
                        glNormal3f(asEdgeNorm[iVertex].fX,   asEdgeNorm[iVertex].fY,   asEdgeNorm[iVertex].fZ);
                        glVertex3f(asEdgeVertex[iVertex].fX, asEdgeVertex[iVertex].fY, asEdgeVertex[iVertex].fZ);
                }
        }
}



//vMarchCube2 performs the Marching Tetrahedrons algorithm on a single cube by making six calls to vMarchTetrahedron
GLvoid vMarchCube2(GLfloat fX, GLfloat fY, GLfloat fZ, GLfloat fScale)
{
        GLint iVertex, iTetrahedron, iVertexInACube;
        GLvector asCubePosition[8];
        GLfloat  afCubeValue[8];
        GLvector asTetrahedronPosition[4];
        GLfloat  afTetrahedronValue[4];

        //Make a local copy of the cube's corner positions
        for(iVertex = 0; iVertex < 8; iVertex++)
        {
                asCubePosition[iVertex].fX = fX + a2fVertexOffset[iVertex][0]*fScale;
                asCubePosition[iVertex].fY = fY + a2fVertexOffset[iVertex][1]*fScale;
                asCubePosition[iVertex].fZ = fZ + a2fVertexOffset[iVertex][2]*fScale;
        }

        //Make a local copy of the cube's corner values
        for(iVertex = 0; iVertex < 8; iVertex++)
        {
                afCubeValue[iVertex] = fSample(asCubePosition[iVertex].fX,
                                                   asCubePosition[iVertex].fY,
                                               asCubePosition[iVertex].fZ);
        }

        for(iTetrahedron = 0; iTetrahedron < 6; iTetrahedron++)
        {
                for(iVertex = 0; iVertex < 4; iVertex++)
                {
                        iVertexInACube = a2iTetrahedronsInACube[iTetrahedron][iVertex];
                        asTetrahedronPosition[iVertex].fX = asCubePosition[iVertexInACube].fX;
                        asTetrahedronPosition[iVertex].fY = asCubePosition[iVertexInACube].fY;
                        asTetrahedronPosition[iVertex].fZ = asCubePosition[iVertexInACube].fZ;
                        afTetrahedronValue[iVertex] = afCubeValue[iVertexInACube];
                }
                vMarchTetrahedron(asTetrahedronPosition, afTetrahedronValue);
        }
}
        

//vMarchingCubes iterates over the entire dataset, calling vMarchCube on each cube
GLvoid vMarchingCubes()
{
        GLint iX, iY, iZ;

        for(iX = 0; iX < mapPoints[0]-1; iX++)
        for(iY = 0; iY < mapPoints[1]-1; iY++)
        for(iZ = 0; iZ < mapPoints[2]-1; iZ++) 
        {
                vMarchCube(iX*fStepSize, iY*fStepSize, iZ*fStepSize, fStepSize);
                //vMarchCube(iX*fStepSize, iY*fStepSize, iZ*fStepSize, fStepSize);
        }
}


// For any edge, if one vertex is inside of the surface and the other is outside of the surface
//  then the edge intersects the surface
// For each of the 4 vertices of the tetrahedron can be two possible states : either inside or outside of the surface
// For any tetrahedron the are 2^4=16 possible sets of vertex states
// This table lists the edges intersected by the surface for all 16 possible vertex states
// There are 6 edges.  For each entry in the table, if edge #n is intersected, then bit #n is set to 1

GLint aiTetrahedronEdgeFlags[16]=
{
        0x00, 0x0d, 0x13, 0x1e, 0x26, 0x2b, 0x35, 0x38, 0x38, 0x35, 0x2b, 0x26, 0x1e, 0x13, 0x0d, 0x00, 
};


// For each of the possible vertex states listed in aiTetrahedronEdgeFlags there is a specific triangulation
// of the edge intersection points.  a2iTetrahedronTriangles lists all of them in the form of
// 0-2 edge triples with the list terminated by the invalid value -1.
//
// I generated this table by hand

GLint a2iTetrahedronTriangles[16][7] =
{
        {-1, -1, -1, -1, -1, -1, -1},
        { 0,  3,  2, -1, -1, -1, -1},
        { 0,  1,  4, -1, -1, -1, -1},
        { 1,  4,  2,  2,  4,  3, -1},

        { 1,  2,  5, -1, -1, -1, -1},
        { 0,  3,  5,  0,  5,  1, -1},
        { 0,  2,  5,  0,  5,  4, -1},
        { 5,  4,  3, -1, -1, -1, -1},

        { 3,  4,  5, -1, -1, -1, -1},
        { 4,  5,  0,  5,  2,  0, -1},
        { 1,  5,  0,  5,  3,  0, -1},
        { 5,  2,  1, -1, -1, -1, -1},

        { 3,  4,  2,  2,  4,  1, -1},
        { 4,  1,  0, -1, -1, -1, -1},
        { 2,  3,  0, -1, -1, -1, -1},
        {-1, -1, -1, -1, -1, -1, -1},
};

// For any edge, if one vertex is inside of the surface and the other is outside of the surface
//  then the edge intersects the surface
// For each of the 8 vertices of the cube can be two possible states : either inside or outside of the surface
// For any cube the are 2^8=256 possible sets of vertex states
// This table lists the edges intersected by the surface for all 256 possible vertex states
// There are 12 edges.  For each entry in the table, if edge #n is intersected, then bit #n is set to 1

GLint aiCubeEdgeFlags[256]=
{
        0x000, 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00, 
        0x190, 0x099, 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90, 
        0x230, 0x339, 0x033, 0x13a, 0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30, 
        0x3a0, 0x2a9, 0x1a3, 0x0aa, 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0, 
        0x460, 0x569, 0x663, 0x76a, 0x066, 0x16f, 0x265, 0x36c, 0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60, 
        0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0x0ff, 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0, 
        0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x055, 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950, 
        0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0x0cc, 0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0, 
        0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc, 0x0cc, 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0, 
        0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x055, 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650, 
        0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc, 0x2fc, 0x3f5, 0x0ff, 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0, 
        0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x066, 0x76a, 0x663, 0x569, 0x460, 
        0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6, 0x0aa, 0x1a3, 0x2a9, 0x3a0, 
        0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c, 0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x033, 0x339, 0x230, 
        0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x099, 0x190, 
        0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x000
};

//  For each of the possible vertex states listed in aiCubeEdgeFlags there is a specific triangulation
//  of the edge intersection points.  a2iTriangleConnectionTable lists all of them in the form of
//  0-5 edge triples with the list terminated by the invalid value -1.
//  For example: a2iTriangleConnectionTable[3] list the 2 triangles formed when corner[0] 
//  and corner[1] are inside of the surface, but the rest of the cube is not.
//
//  I found this table in an example program someone wrote long ago.  It was probably generated by hand

GLint a2iTriangleConnectionTable[256][16] =  
{
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
        {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
        {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
        {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
        {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
        {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
        {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
        {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
        {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
        {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
        {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
        {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
        {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
        {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
        {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
        {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
        {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
        {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
        {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
        {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
        {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
        {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
        {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
        {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
        {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
        {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
        {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
        {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
        {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
        {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
        {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
        {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
        {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
        {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
        {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
        {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
        {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
        {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
        {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
        {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
        {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
        {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
        {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
        {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
        {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
        {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
        {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
        {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
        {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
        {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
        {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
        {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
        {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
        {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
        {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
        {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
        {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
        {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
        {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
        {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
        {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
        {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
        {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
        {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
        {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
        {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
        {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
        {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
        {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
        {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
        {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
        {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
        {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
        {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
        {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
        {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
        {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
        {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
        {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
        {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
        {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
        {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
        {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
        {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
        {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
        {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
        {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
        {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
        {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
        {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
        {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
        {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
        {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
        {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
        {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
        {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
        {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
        {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
        {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
        {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
        {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
        {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
        {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
        {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
        {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
        {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
        {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
        {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
        {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
        {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
        {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
        {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
        {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
        {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
        {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
        {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
        {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
        {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
        {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
        {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
        {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
        {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
        {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
        {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
        {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
        {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
        {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
        {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
        {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
        {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
        {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
        {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
        {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
        {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
        {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
        {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
        {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
        {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
        {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
        {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
        {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
        {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
        {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
        {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
        {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
        {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
        {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
        {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
        {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
        {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
        {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
        {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
        {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
        {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};
